use crate::protocol::{
    cmd_advanced_mode_settings, cmd_print_information_command, cmd_raster_graphics_transfer_line,
    cmd_specify_margin_amount, cmd_switch_automatic_status_notification_code,
    cmd_various_mode_settings, parse_status_response, AdvancedModeSettings, MediaType,
    StatusResponse, StatusType, VariousModeSettings, CMD_INITIALISE, CMD_INVALIDATE, CMD_PRINT,
    CMD_PRINT_WITH_FEEDING, CMD_SELECT_COMPRESSION_MODE_UNCOMPRESSED,
    CMD_STATUS_INFORMATION_REQUEST, CMD_SWITCH_DYNAMIC_COMMAND_MODE_RASTER,
};
use eyre::Context;
use std::time::Duration;
use tracing::debug;

pub trait P710btDevice {
    /// Transmits a frame to the device.
    fn transmit(&mut self, bytes: &[u8]) -> eyre::Result<()>;

    /// Receives a frame from the device if possible, but will not block to read it.
    fn receive_nonblock(&mut self, max: usize) -> eyre::Result<Option<Vec<u8>>>;

    /// Receives a frame from the device, blocking to wait for one if necessary.
    fn receive_block(&mut self, max: usize) -> eyre::Result<Vec<u8>>;
}

pub struct P710btCommander<D: P710btDevice> {
    device: D,
    last_status: Option<StatusResponse>,
}

impl<D: P710btDevice> P710btCommander<D> {
    pub fn new(device: D) -> eyre::Result<Self> {
        Ok(Self {
            device,
            last_status: None,
        })
    }

    pub fn invalidate_and_initialise(&mut self) -> eyre::Result<()> {
        self.device.transmit(CMD_INVALIDATE).context("invalidate")?;
        self.device.transmit(CMD_INITIALISE).context("initialise")?;
        Ok(())
    }

    pub fn status(&mut self) -> eyre::Result<StatusResponse> {
        self.device
            .transmit(CMD_STATUS_INFORMATION_REQUEST)
            .context("status")?;
        let response = self.device.receive_block(32).context("status (R)")?;
        let status_response =
            parse_status_response(&response).context("failed to parse status response")?;
        self.last_status = Some(status_response.clone());
        debug!("polled status: {status_response:?}");
        Ok(status_response)
    }

    pub fn print_information(
        &mut self,
        media_type: MediaType,
        media_width_mm: u8,
        raster_number_lines: u32,
        first_page: bool,
    ) -> eyre::Result<()> {
        self.device
            .transmit(&cmd_print_information_command(
                media_type,
                media_width_mm,
                raster_number_lines,
                first_page,
            ))
            .context("PIC")?;
        Ok(())
    }

    pub fn set_raster_mode_and_automatic_notification(&mut self) -> eyre::Result<()> {
        self.device
            .transmit(CMD_SWITCH_DYNAMIC_COMMAND_MODE_RASTER)
            .context("set raster mode")?;

        self.device
            .transmit(&cmd_switch_automatic_status_notification_code(true))
            .context("set notify mode")?;

        self.device
            .transmit(CMD_SELECT_COMPRESSION_MODE_UNCOMPRESSED)
            .context("set uncompressed")?;

        Ok(())
    }

    pub fn set_mode_settings(
        &mut self,
        vms: VariousModeSettings,
        ams: AdvancedModeSettings,
        margin_dots: u16,
    ) -> eyre::Result<()> {
        self.device
            .transmit(&cmd_various_mode_settings(vms))
            .context("various mode settings")?;
        self.device
            .transmit(&cmd_advanced_mode_settings(ams))
            .context("advanced mode settings")?;
        self.device
            .transmit(&cmd_specify_margin_amount(margin_dots))
            .context("margins")?;
        Ok(())
    }

    pub fn raster_line(&mut self, line: [u8; 16]) -> eyre::Result<Option<StatusResponse>> {
        self.device
            .transmit(&cmd_raster_graphics_transfer_line(line))
            .context("raster line")?;
        if let Some(status_notification) = self
            .device
            .receive_nonblock(32)
            .context("raster line read (opt)")?
        {
            let status_response = parse_status_response(&status_notification)
                .context("failed to parse notification")?;
            debug!("status notification (L): {status_response:?}");
            self.last_status = Some(status_response.clone());
            return Ok(Some(status_response));
        }
        Ok(None)
    }

    pub fn finish_printing(&mut self, last_page: bool) -> eyre::Result<()> {
        let tx = if last_page {
            CMD_PRINT_WITH_FEEDING
        } else {
            CMD_PRINT
        };
        self.device.transmit(tx).context("finish page")?;

        loop {
            if let Some(ref status) = self.last_status {
                if status.status_type == Some(StatusType::PhaseChange) && status.reception_possible
                {
                    break;
                }
            }
            std::thread::sleep(Duration::from_millis(50));
            if let Some(status_notification) = self
                .device
                .receive_nonblock(32)
                .context("wait for finish")?
            {
                let status_response = parse_status_response(&status_notification)
                    .context("failed to parse notification")?;
                debug!("status notification (F): {status_response:?}");
                self.last_status = Some(status_response.clone());
            }
        }

        Ok(())
    }
}
