use bitflags::bitflags;
use eyre::bail;
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;

//// Invalidate
pub const CMD_INVALIDATE: &[u8] = &[0x00; 100];

//// Initialise
pub const CMD_INITIALISE: &[u8] = &[0x1B, 0x40];

//// Status information request
pub const CMD_STATUS_INFORMATION_REQUEST: &[u8] = &[0x1B, 0x69, 0x53];

bitflags! {
    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct ErrorInformation: u16 {
        const NO_MEDIA = 0x0001;
        const CUTTER_JAM = 0x0004;
        const WEAK_BATTERIES = 0x0008;
        const HIGH_VOLTAGE_ADAPTER = 0x0040;
        const REPLACE_MEDIA = 0x0100;
        const COVER_OPEN = 0x1000;
        const OVERHEATING = 0x2000;
    }
}

#[derive(FromPrimitive, Copy, Clone, Debug)]
pub enum MediaType {
    LaminatedTape = 0x01,
    NonLaminatedTape = 0x03,
    HeatShrinkTube2To1 = 0x11,
    HeatShrinkTube3To1 = 0x17,
    Incompatible = 0xFF,
}

#[derive(FromPrimitive, Copy, Clone, Debug, Eq, PartialEq)]
pub enum StatusType {
    ReplyToStatusRequest = 0x00,
    PrintingCompleted = 0x01,
    ErrorOccurred = 0x02,
    ExitIfModeUnused = 0x03,
    TurnedOff = 0x04,
    Notification = 0x05,
    PhaseChange = 0x06,
}

#[derive(FromPrimitive, Copy, Clone, Debug)]
pub enum TapeColour {
    White = 0x01,
    Other = 0x02,
    // TODO
}

#[derive(FromPrimitive, Copy, Clone, Debug)]
pub enum TextColour {
    White = 0x01,
    Red = 0x04,
    Blue = 0x05,
    Black = 0x08,
    // TODO
}

#[derive(Copy, Clone, Debug)]
pub struct TapeSizeInfo {
    /// 'Width' of the tape in millimetres
    pub mm: u32,

    /// Practical print height limit on the tape.
    /// Based on empirical measurements.
    pub practical_px_limit: u32,

    /// Specified print height in dots from the protocol datasheet.
    pub datasheet_px_limit: u32,
}

/// Tape size (mm) to pixels high that can be printed in practice, then the number that can be printed according to the datasheet
pub const PIXELS_ON_TAPE_SIZES: &[TapeSizeInfo] = &[
    TapeSizeInfo {
        mm: 12,
        // Actually 81 but it seems like an awkward number!
        practical_px_limit: 80,
        // 70 dots is way too conservative?
        datasheet_px_limit: 70,
    },
    TapeSizeInfo {
        mm: 24,
        // the printer can only print 128 dots; can't push our luck
        practical_px_limit: 128,
        datasheet_px_limit: 128,
    },
];

#[derive(Clone, Debug)]
pub struct StatusResponse {
    pub error_information: ErrorInformation,
    /// Media width in mm. 0 = no media. 4 = 3.5 mm.
    pub media_width_mm: u8,
    pub media_type: Option<MediaType>,
    pub status_type: Option<StatusType>,
    pub tape_colour: Option<TapeColour>,
    pub text_colour: Option<TextColour>,
    pub reception_possible: bool,
}

pub fn parse_status_response(input: &[u8]) -> eyre::Result<StatusResponse> {
    if input.len() != 32 {
        bail!("status response should be 32 bytes");
    }

    let error_information =
        ErrorInformation::from_bits_truncate(((input[9] as u16) << 8) | (input[8] as u16));

    let media_type = FromPrimitive::from_u8(input[11]);
    let status_type = FromPrimitive::from_u8(input[18]);
    let tape_colour = FromPrimitive::from_u8(input[24]);
    let text_colour = FromPrimitive::from_u8(input[25]);
    let reception_possible = input[19] == 0;

    Ok(StatusResponse {
        error_information,
        media_width_mm: input[10],
        media_type,
        status_type,
        tape_colour,
        text_colour,
        reception_possible,
    })
}

// Switch dynamic command mode
// Don't bother implementing the other modes: as the manual says, 'be sure to switch to this mode'.
pub const CMD_SWITCH_DYNAMIC_COMMAND_MODE_RASTER: &[u8] = &[0x1B, 0x69, 0x61, 0x01];

//// Switch automatic status notification code
pub fn cmd_switch_automatic_status_notification_code(notify: bool) -> [u8; 4] {
    [0x1B, 0x69, 0x21, if notify { 0x00 } else { 0x01 }]
}

//// Print information command

/// Generate the 'print information command'.
/// `raster_number` is poorly documented but seems to be the number of raster commands to be issued
/// (?) or the number of lines. I'm not entirely sure which.
pub fn cmd_print_information_command(
    media_type: MediaType,
    media_width_mm: u8,
    raster_number: u32,
    first_page: bool,
) -> [u8; 13] {
    let mut out = [
        0x1B,
        0x69,
        0x7A,
        0x02 | 0x04 | 0x80,
        media_type as u8,
        media_width_mm,
        0,
        0,
        0,
        0,
        0,
        if first_page { 0 } else { 1 },
        0,
    ];

    out[7..11].copy_from_slice(&raster_number.to_le_bytes());

    out
}

//// Various mode settings
bitflags! {
    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct VariousModeSettings: u8 {
        const AUTO_CUT = 0x40;
        const MIRROR_PRINTING = 0x80;
    }
}

pub fn cmd_various_mode_settings(n1: VariousModeSettings) -> [u8; 4] {
    [0x1B, 0x69, 0x4D, n1.bits()]
}

//// Specify the page number (NOT FOR P710BT)
// Not implemented: not for P710BT.

//// Advanced mode settings
bitflags! {
    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub struct AdvancedModeSettings: u8 {
        // 0x01, 0x02 unused
        // 0x04: half cut (not supported by P710BT)
        const NO_CHAIN_PRINTING = 0x08;
        const SPECIAL_TAPE_NO_CUTTING = 0x10;
        // 0x20 unused.
        const HIGH_RES_PRINTING = 0x40;
        const NO_BUFFER_CLEARING_WHEN_PRINTING = 0x80;
    }
}

pub fn cmd_advanced_mode_settings(n1: AdvancedModeSettings) -> [u8; 4] {
    [0x1B, 0x69, 0x4B, n1.bits()]
}

//// Specify margin amount (feed amount)
pub fn cmd_specify_margin_amount(num_dots: u16) -> [u8; 5] {
    let num_dots_le = num_dots.to_le_bytes();
    [0x1B, 0x69, 0x64, num_dots_le[0], num_dots_le[1]]
}

//// Select compression mode
// Don't bother with the other modes for now.
pub const CMD_SELECT_COMPRESSION_MODE_UNCOMPRESSED: &[u8] = &[0x4D, 0x00];

//// Raster graphics transfer
pub fn cmd_raster_graphics_transfer_line(line: [u8; 16]) -> [u8; 19] {
    let mut out = [0; 19];
    // manual suggests n=16 hardcoded when using uncompressed mode
    out[0..3].copy_from_slice(&[0x47, 16, 0x00]);
    out[3..].copy_from_slice(&line);
    out
}

//// Zero raster graphics
// Not bothering for now.

//// Print command
pub const CMD_PRINT: &[u8] = &[0x0C];

//// Print command with feeding
pub const CMD_PRINT_WITH_FEEDING: &[u8] = &[0x1A];
