use crate::commander::P710btDevice;
use eyre::{bail, Context};
use rusb::{Device, DeviceHandle, DeviceList, GlobalContext};
use std::time::Duration;
use tracing::{debug, info, warn};

pub const VENDOR_BROTHER: u16 = 0x04F9;
pub const PRODUCT_P710BT: u16 = 0x20AF;

pub struct UsbP710bt {
    pub device_handle: DeviceHandle<GlobalContext>,
}

impl UsbP710bt {
    pub fn list_devices() -> eyre::Result<Vec<Device<GlobalContext>>> {
        let mut result = Vec::new();

        let device_list = DeviceList::new()?;
        for device in device_list.iter() {
            match device.device_descriptor() {
                Ok(descriptor) => {
                    debug!(
                        "found USB device {:04x}:{:04x}",
                        descriptor.vendor_id(),
                        descriptor.product_id()
                    );
                    if descriptor.product_id() == PRODUCT_P710BT
                        && descriptor.vendor_id() == VENDOR_BROTHER
                    {
                        info!("found P710BT");
                        result.push(device);
                    }
                }
                Err(err) => {
                    warn!(
                        "Ignoring USB device {device:?} as can't read device descriptor: {err:?}."
                    );
                }
            }
        }

        Ok(result)
    }

    pub fn open(device: Device<GlobalContext>) -> eyre::Result<Self> {
        let mut device_handle = device.open().context("failed to open device")?;
        if device_handle
            .kernel_driver_active(0)
            .context("check kernel driver active")?
        {
            device_handle
                .detach_kernel_driver(0)
                .context("detach kernel driver")?;
        }
        device_handle
            .claim_interface(0)
            .context("claim usb interface")?;
        Ok(UsbP710bt { device_handle })
    }
}

// n.b. the top bit is set because it's an IN endpoint. This seems to be a USB standard.
// The manual calls this endpoint 1.
const PRINTER_TO_HOST_ENDPOINT: u8 = 0x81;
const HOST_TO_PRINTER_ENDPOINT: u8 = 0x02;

impl P710btDevice for UsbP710bt {
    fn transmit(&mut self, bytes: &[u8]) -> eyre::Result<()> {
        let written = self
            .device_handle
            .write_bulk(HOST_TO_PRINTER_ENDPOINT, bytes, Duration::from_secs(0))
            .context("usb write")?;
        if written < bytes.len() {
            bail!(
                "failed to write whole buffer ({} written of {}) to usb",
                written,
                bytes.len()
            );
        }
        Ok(())
    }

    fn receive_nonblock(&mut self, max: usize) -> eyre::Result<Option<Vec<u8>>> {
        let mut buf = vec![0u8; max];
        let mut read_tot = 0;

        while read_tot < max {
            match self.device_handle.read_bulk(
                PRINTER_TO_HOST_ENDPOINT,
                &mut buf[read_tot..],
                Duration::from_millis(if read_tot == 0 { 1 } else { 0 }),
            ) {
                Ok(just_read) => {
                    if just_read == 0 {
                        if read_tot > 0 {
                            bail!("partial usb read; have {} of {} bytes read", read_tot, max);
                        } else {
                            return Ok(None);
                        }
                    }
                    read_tot += just_read;
                }
                Err(rusb::Error::Timeout) => {
                    if read_tot > 0 {
                        bail!(
                            "UNREACHABLE TIMEOUT after partial usb read; have {} of {} bytes read",
                            read_tot,
                            max
                        );
                    } else {
                        return Ok(None);
                    }
                }
                Err(other) => {
                    bail!("usb read: {other:?}");
                }
            }
        }

        Ok(Some(buf))
    }

    fn receive_block(&mut self, max: usize) -> eyre::Result<Vec<u8>> {
        let mut buf = vec![0u8; max];
        let mut read_tot = 0;

        while read_tot < max {
            let just_read = self
                .device_handle
                .read_bulk(
                    PRINTER_TO_HOST_ENDPOINT,
                    &mut buf[read_tot..],
                    Duration::from_millis(0),
                )
                .context("usb read")?;
            if just_read == 0 {
                bail!("empty usb read; have {} of {} bytes read", read_tot, max);
            }
            read_tot += just_read;
        }

        Ok(buf)
    }
}
